/* Texture atlasses: 0 = uncompressed, 1 = compressed */
uniform lowp sampler2DArray textures[2];

uniform mediump usampler2D indirectionTexture;

/* Texture streaming tile visualization */
//#define VISUALIZE_TEXTURE_STREAMING

uniform highp usampler2D textureBounds;

#ifdef VISUALIZE_TEXTURE_STREAMING
lowp vec4 randomColor(highp uint id) {
    lowp vec4 c;
    highp float fid = float(id);
    c.r = mod(fid*2.4, 33.0)/33.0;
    c.g = mod(fid*7.0, 17.0)/17.0;
    c.b = mod(fid*13.0, 21.0)/21.0;
    c.a = 1.0;
    return c;
}
const vec4 BLACK = vec4(0, 0, 0, 1.0);
const vec4 WHITE = vec4(1.0);
const vec4 PINK = vec4(1.0, 0.0, 1.0, 1.0);
#endif

highp float fastInv2Exp(lowp int exp) {
    lowp int i = 1 << exp;
    return 1.0/float(i);
    // TODO: This implementation seems to only work on some
    // platforms:
    //return uintBitsToFloat((uint(-exp + 127) & 0x7Fu) << 23u);
}

lowp int fastIntLog2(highp float v) {
    mediump uint bits = floatBitsToUint(v) >> 23u;
    return int(bits & 255u) - 127;
}

lowp int mipMapLevel(vec2 pixelCoords, lowp int maxMip) {
    highp vec2 dx = dFdx(pixelCoords);
    highp vec2 dy = dFdy(pixelCoords);

    /* Equirectangular sampling may cause 0 here, which
     * causes the wrong mip level to get selected under rare
     * circumstances, causing a very visible seam. */
    dx = dx*dx;
    dx.x = max(0.0001, dx.x);

    highp float d = max(dx.x + dx.y, dot(dy, dy));

    /* Clamp to valid mip range */
    return clamp(fastIntLog2(d) >> 1, 0, maxMip);
}

mediump uint sampleTile(lowp int lod, mediump uvec2 offset, vec2 uv) {
    float toIndirection = fastInv2Exp(lod + 8);
    vec2 offs = ceil(vec2(offset)*toIndirection);
    mediump ivec2 indirectionSt = ivec2(offs + uv*toIndirection);
    return texelFetch(indirectionTexture, indirectionSt, lod).r;
}

vec4 tileStack(lowp sampler2DArray stack, mediump uvec2 offset, mediump uvec2 size, vec2 uv, lowp int lod) {
    /* Texture coordinates on virtual texture atlas */
    vec2 wrappedSt = fract(uv)*vec2(size);

    mediump uint tile = sampleTile(lod, offset, wrappedSt);

    lowp int tileLod = int(tile >> 11u);
    /* Max tile cache is 2048 tiles */
    tile = tile & 0x7FFu;

    /* Compute UVs relative to the tile */
    highp float ti = fastInv2Exp(8 + lod + tileLod);
    highp vec2 tileUV = fract(wrappedSt*ti);

    /* Bilinear filtering */
    const float twoPixels = 2.0/260.0;
    const float rescale = 256.0/260.0;
    lowp vec4 c = textureLod(stack, vec3(tileUV*rescale + twoPixels, float(tile)), 0.0);

#ifdef VISUALIZE_TEXTURE_STREAMING
#define LINE_WIDTH 2.0*twoPixels
#define PAD twoPixels
    if(tileUV.x < PAD + LINE_WIDTH || tileUV.x > 1.0 - PAD - LINE_WIDTH
        || tileUV.y < PAD + LINE_WIDTH || tileUV.y > 1.0 - PAD - LINE_WIDTH)
        return vec4(randomColor(tile).rgb, 0.4);
#endif
    return c;
}

vec4 tileStack(lowp sampler2DArray stack, mediump uvec2 offset, mediump uvec2 size, vec2 uv) {
    lowp int maxLod = fastIntLog2(float(max(size.x, size.y))) - 8;
    vec2 st = uv*vec2(size);
    lowp int lod = mipMapLevel(st, maxLod);
    return tileStack(stack, offset, size, uv, lod);
}

bool unpackBounds(mediump uint t, out lowp uint layer, out mediump uvec4 bounds) {
    highp uvec4 boundsPacked = texelFetch(textureBounds,
        ivec2(int(t & 0x3FFu), int(t >> 10u)), 0);
    layer = boundsPacked.x >> 8u;
    if(layer == 0xFFu) {
        /* Compressed image bounds */
        bounds = uvec4(
            (boundsPacked.y & 0xFFu) << 8u,
            boundsPacked.y & 0xFF00u,
            boundsPacked.z,
            boundsPacked.w
        );
        return true;
    }
    /* Uncompressed image bounds */
    bounds = uvec4(
        boundsPacked.y & 0xFFFu,
        ((boundsPacked.x & 0xFFu) << 4u) | (boundsPacked.y >> 12u),
        boundsPacked.z,
        boundsPacked.w
    );
    return false;
}

vec4 textureAtlas(vec2 uv, mediump uvec4 bounds, lowp uint layer, bool compressed) {
    if(compressed) {
        return tileStack(textures[1], bounds.xy, bounds.zw, uv);
    } else {
        vec3 finalUV = vec3(vec2(bounds.xy) + fract(uv)*vec2(bounds.zw), layer >> 1u)/
            vec3(textureSize(textures[0], 0).xy, 1);
        return texture(textures[0], finalUV);
    }
}

vec4 textureAtlas(mediump uint t, vec2 uv) {
    lowp uint layer = 0u;
    mediump uvec4 bounds;
    bool compressed = unpackBounds(t, layer, bounds);
    return textureAtlas(uv, bounds, layer, compressed);
}

vec4 textureAtlas(mediump uint t, vec2 uv, lowp int mip) {
    lowp uint layer = 0u;
    mediump uvec4 bounds;
    if(unpackBounds(t, layer, bounds)) {
        /* Get the bounds on the texture altas */
        return tileStack(textures[1], bounds.xy, bounds.zw, uv, mip);
    } else {
        vec3 finalUV = vec3(vec2(bounds.xy) + fract(uv)*vec2(bounds.zw), layer >> 1u)/
            vec3(textureSize(textures[0], 0).xy, 1);
        return texture(textures[0], finalUV);
    }
}

#ifdef NORMAL_MAPPING
vec3 normalMapping(mediump vec3 normalizedNormal, mediump uint t) {
    mediump vec3 normalizedTangent = normalize(fragTangent.xyz);
    mediump mat3 tbn = mat3(
        normalizedTangent,
        normalize(cross(normalizedNormal, normalizedTangent)*fragTangent.w),
        normalizedNormal);
    mediump vec3 normal = tbn*(normalize((textureAtlas(t, fragTextureCoords).rgb*2.0 - vec3(1.0))));
    return normal;
}
#endif
